﻿/*
 *  @copyright 2019 Joe McShea
 *  @license under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 */
(function() {
    // this is the only thing you need to modify to add new fields, it is an array of
    // field internal names
    var fields = [
        "SingleTag",
        "TagsCovered",
        "TagsNotCovered"
    ];

    /*********************************************************************************
     * If you just want to use this CSR for one or more fields, just modify the array
     * above and leavs alone the code below. Disclaimer: technically, I do not recommend
     * using scripts in production that you don't understand, so if you're up for it read
     * the code below, but there is no need to modify it unless you're good at JavaScript
     * and want to adjust the way it works.
     *********************************************************************************/
     
    // test is form with client side rendering
    if (typeof(SPClientTemplates) === 'undefined')
        return;
    
    /*
     * Implementation class for the overrides.
     */
    var impl = {
        fieldMap: {},
        cssLoaded: false,

        /*
         * Implementation for the display form and view overrides.
         */
        renderDisplay: function(ctx) {
            var values = ctx.CurrentItem[ctx.CurrentFieldSchema.Name];

            // create the outer div
            var div = document.createElement("div");
            if (values.length > 0) {
                // split the values into an array (on semi-colon)
                if (typeof(values) === "string") {
                    values = values.split(";");
                }

                // for each value
                values.forEach(function (v) {
                    // create an entity span and append it to the div
                    var span = document.createElement("span");
                    span.className = "csrdemos-entity";
                    span.innerHTML = v;
                    div.appendChild(span);
                });
            }

            return div.outerHTML;
        },

        /*
         * Implementation for the new and edit form override.
         */
        renderEdit: function(ctx) {
            var current = SPClientTemplates.Utility.GetFormContextForCurrentField(ctx);

            // add a field map entry for the field, and set it's source (the autocomplete list)
            // and schema (the field schema from the SharePoint form context)
            var f = impl.fieldMap[current.fieldName] = {};
            f.source = current.fieldSchema.MultiChoices;
            f.schema = current.fieldSchema;

            // construct the outer html for our control
            var outerDiv = impl.appendOuterContainer(current);
            var entityEditor = outerDiv.querySelector(".csrdemos-entityeditor");

            // initialize the editor with current values
            impl.appendInitialEntities(ctx, current, entityEditor);

            // add the text input to the entitiy edity div
            impl.appendInput(ctx, current, entityEditor);

            // register a callback to return the current value
            current.registerGetValueCallback(current.fieldName, function() {
                return impl.getFieldValue(current);
            });

            // register validators for this control
            impl.registerValidators(current);

            return outerDiv.outerHTML;
        },
        
        /*
         * Implementation for the pre render override. It's purpose is to inject CSS.
         */
        preRender: function(ctx) {
            if (!impl.cssLoaded) {
                impl.cssLoaded = true;
                var head = document.querySelector("head");

                // create a new style element with our CSS
                var style = document.createElement("style");
                style.setAttribute("type", "text/css");
                style.innerHTML = entityEditorCss;

                // append the style element to the head
                head.appendChild(style);

                // append the awesomecomplete CSS to the head, you can't load CSS through JSLink, 
                // and while I could load it in a ScriptBlock custom action, that would load it 
                // on every page and we dont need that
                var awesomeCompleteCss = document.createElement("link");
                awesomeCompleteCss.setAttribute("rel", "stylesheet");
                awesomeCompleteCss.setAttribute("href", _spPageContextInfo.siteAbsoluteUrl + "/Style Library/awesomplete.css");
                head.appendChild(awesomeCompleteCss);
            }
        },
        
        /*
         * Implementation for the post render override. It's purpose is to attach event
         * receivers.
         */
        postRender: function(ctx) {
            // get the field name from the schema, there's only one field in the schema array
            // at this point, so that's the one we want
            var fieldName = ctx.ListSchema.Field[0].Name;

            // if this field is in our array, set up event handlers on it
            if (fields.indexOf(fieldName) > -1) {
                // find the entity editor and input from the field name
                var entityEditor = document.getElementById(fieldName + 'EntityEditor');
                var entityEditorInput = entityEditor.querySelector("input");

                // get the field resources from the map and the autocomplete source
                var f = impl.fieldMap[fieldName];
                var source = f.source;

                // initialize the autocomplete on the input
                f.awesomplete = new Awesomplete(entityEditorInput, {
                    minChars: 1,
                    autoFirst: true,
                    list: source,
                    filter: Awesomplete.FILTER_CONTAINS,
                    maxItems: 30,
                    sort: function (a, b) {
                        // custom compare method sorts alphabetically (case insensitive)
                        if (a.toLowerCase() < b.toLowerCase()) return -1;
                        if (a.toLowerCase() > b.toLowerCase()) return 1;
                        return 0;
                    }
                });

                // get called back after the autocomplete has selected an item
                entityEditorInput.addEventListener("awesomplete-selectcomplete", impl.handleSelectComplete);

                // also add a key down even callback
                entityEditorInput.addEventListener("keydown", impl.handleInputKeyDown);

                // if the outer div receives a click event, put focus on the input
                entityEditor.addEventListener("click", impl.handleEditorClick);

                // add remove handlers to the little x anchor of each entity
                var entities = entityEditor.querySelectorAll(".csrdemos-remove");
                for(var i=0; i<entities.length; i++) {
                    entities[i].addEventListener("click", impl.handleEntityRemove);
                }
            }
        },

        
        /*
         * Setup validation for the input form.
         */
        registerValidators: function(current) {
            // create a validator set
            var fieldValidators = new SPClientForms.ClientValidation.ValidatorSet();
 
            // create a custom validator with an object literal insead of new and a constructor,
            // this creates a closure where I can still access the form context (current), otherwise
            // the validate method only gets value passed to it and I need more than that to 
            // validate
            fieldValidators.RegisterValidator({
                Validate: function(value) {
                    var isError = false;
                    var errorMessage = '';

                    // if fill in choice is enabled for this field
                    if (current.fieldSchema.FillInChoice === false) {
                        // get the entity editor and inpu
                        var entityEditor = document.getElementById(current.fieldSchema.Name + 'EntityEditor');
                        var entityInput = entityEditor.querySelector("input");
                        // the getValue callback is called before validate, and will resolve anything it can and
                        // clear the input, so if there is anthing left in the input, it's invalid
                        if (entityInput.value.length > 0) {
                            isError = true;
                            errorMessage = "'" + entityInput.value + 
                                "' is not resolved; Fill in choices are not support, so all entities must be resolved.";
                        }
                    }

                    // return the validation result
                    return new SPClientForms.ClientValidation.ValidationResult(isError, errorMessage);
                }
            });

            // if required, add a required field validator
            if (current.fieldSchema.Required) {
                fieldValidators.RegisterValidator(new SPClientForms.ClientValidation.RequiredValidator());
            }

            // register a callback method for the validators, to render an error message
            current.registerValidationErrorCallback(current.fieldName, function(error) {
                var entityError = document.getElementById(current.fieldName + 'EntityEditorError');
                entityError.setAttribute('role', error.validationError ? 'Alert' : '');
                entityError.innerHTML = error.errorMessage;
            });

            // register the validator set, this replaces any previous validators, so if I don't include
            // a required validator, the field isn't required
            current.registerClientValidator(current.fieldName, fieldValidators);
        },

        /*
         * Return the current value from the data-value attribute of my div.
         */
        getFieldValue: function(current) {
            var result = [];

            // get the field name from the form context
            var fieldName = current.fieldName;

            // get the field resources from the map, and from them the autocomplete source and schema
            var f = impl.fieldMap[fieldName];
            var source = f.source;
            var schema = f.schema;

            // get the entity editor, the input, and the validation error message span
            var entityEditor = document.getElementById(current.fieldName + 'EntityEditor');
            var entityEditorInput = entityEditor.querySelector("input");
            var entityError = document.getElementById(current.fieldName + 'EntityEditorError');

            // if there is an unresolved value in the input, see if it matches a valid choice (case insensative)
            if (entityEditorInput.value.length > 0) {
                var index = impl.inArrayIgnoreCase(entityEditorInput.value, source);
                if (index > -1) {
                    // if we found a valid choice, add it as an entity, clear the input, and clear the validation
                    // error if any, use the source array instead of the input.value so we add the choice with
                    // the correct case
                    impl.appendEntity(
                        current.fieldName, 
                        source[index], 
                        entityEditorInput);
                        
                    entityError.setAttribute('role', '');
                    entityError.innerHTML = "";
                }
            }

            // if there is still an unresolved value in input, and fill in choices are allowed, add whatever
            // is in input as an entity
            if (schema.FillInChoice === true) {
                if (entityEditorInput.value.length > 0) {
                    impl.appendEntity(current.fieldName, entityEditorInput.value, entityEditorInput);
                    entityError.setAttribute('role', '');
                    entityError.innerHTML = "";
                }
            }

            // now scoop up all of the entities (all the span.csrdemos-entity elements)
            var entities = entityEditor.querySelectorAll('.csrdemos-entity');
            if(entities !== null)
            {
                for(var i=0; i<entities.length; i++) {
                    result.push(entities[i].getAttribute("data-value"));
                }
            }

            // if this is a multi-choice field, join the array and format as a multi-choice value
            if (schema.FieldType === "MultiChoice") {
                if (result.length === 0) {
                    result = '';
                } else {
                    result = ';#' + result.join(';#') + ';#';
                }
            }
            // otherwise it's a single choice, we don't have to worry about finding multiple entities because the
            // input gets hidden for single choice any time we have an entity
            else {
                result = result.length === 1 ? result[0] : '';
            }

            // return the result as a string, if anything is still in the input at this point (i.e. there is an
            // unresolved entity), it is a validation error which will be handled by the custom validator
            return result;
        },

        /*
         * Handle the click event on the little x anchor for an entity, which removes the entity.
         */
        handleEntityRemove: function (e) {
            // don't bubble up or do the default action, I want to handle this event completely
            e.preventDefault();
            e.stopPropagation();

            // get the anchor tag that was clicked and the field name from it
            var anchor = e.target;
            var fieldName = anchor.getAttribute("data-fieldname");

            // get our field resources from the map, and the source list for autocomplete
            var f = impl.fieldMap[fieldName];
            var source = f.source;

            // get the entity editor and input from the anchor tag
            var entityEditor = anchor.parentNode.parentNode;
            var entityEditorInput = entityEditor.querySelector("input.csrdemos-entityeditorinput");
            
            // remove the event listener for the x anchor
            anchor.removeEventListener("click", impl.handleEntityRemove);
            
            // remove the span for this value, and it back to the source array to be used for autocomplete
            source.push(anchor.getAttribute("data-value"));
            entityEditor.removeChild(anchor.parentNode); 

            // re-initialize autocomplete
            f.awesomplete.list = source;
            
            // show the input, in case it's a single entry input that was hidden due to being 'full'
            entityEditorInput.style.display = "inline-block";
            entityEditorInput.focus();
        },

        /*
         * Handle the key down event for the input. If the key is a return, but the input value
         * is empty, it does nothing. If it has a value in the input, it trys
         * to resolve the input to an entity (equals ignoring case one of the choices). If it can't
         * resolve it, but the field allows fill in choices, then it adds whatever is in the input to 
         * the fill in choices. If fill in choices is not enabled, it shows a validation error message.
         * And finally, if any other key is pressed, it clears any existing validation
         * message. Also, if the input is empty, and they hit the backspace key, remove the last
         * resolved entity if there are any.
         */
        handleInputKeyDown: function (e) {
            // get the entity editor, input and field name
            var entityEditorInput = e.target;
            var entityEditor = entityEditorInput.parentNode.parentNode;
            var fieldName = entityEditor.getAttribute("data-fieldname");

            // get the field resources from the map, and from them the autocomplete source and field schema
            var f = impl.fieldMap[fieldName];
            var schema = f.schema;
            var source = f.source;

            // backspace key
            if (e.which === 8) {
                // if the input is empty
                if (entityEditorInput.value.length === 0) {
                    var entities = entityEditor.querySelectorAll(".csrdemos-entity");
                    // if we have entities
                    if (entities.length > 0) {
                        // find the last entity
                        var last = entities[entities.length - 1];

                        // remove the event listener for the x anchor
                        last.removeEventListener("click", impl.handleEntityRemove);

                        // remove the entity span
                        var value = last.getAttribute("data-value");
                        entityEditor.removeChild(last);
                        if (source.indexOf(value) < 0) {
                            // push the value back into the source list
                            source.push(value);
                        }
                    }
                }
            }

            // if the key is return
            var entityError = document.getElementById(fieldName + 'EntityEditorError');
            if (e.which === 13) {
                e.preventDefault();
                var val = entityEditorInput.value;
                
                // if there is nothing in the input, do nothing
                if (val.length > 0) {
                    var index = impl.inArrayIgnoreCase(val, source);

                    // if the input text equals, ignoring case, a source item, add the source item
                    if (index > -1) {
                        impl.appendEntity(fieldName, source[index], entityEditor);
                        f.awesomplete.close();
                    } else if (!impl.isDuplicateValue(entityEditor, val)) { // ignore duplicate entries
                        // if fill-in choices are allowed, add whatever they typed
                        if (schema.FillInChoice === true) {
                            impl.appendEntity(fieldName, val, entityEditor);
                            f.awesomplete.close();
                        }
                        else { // else raise a validation error
                            var errorMessage = "'" + val +
                                "' is not resolved; Fill in choices are not support, so all entities must be resolved.";
                            entityError.setAttribute('role', 'alert');
                            entityError.innerHTML = errorMessage;
                        }
                    } 
                }
            } 
            else {
                // on any other key, just clear the error span, since we won't know if 
                // it's an error until they try to resolve it again (i.e. hit return or submit)
                entityError.setAttribute('role', '');
                entityError.innerHTML = "";
            }        
        },

        /*
         * If the user clicks anywhere inside the entity editor, focus the input. Otherwise,
         * since the input itself has no border, it could be confusing to the user.
         */
        handleEditorClick: function(e) {
            var fieldName = e.target.getAttribute("data-fieldname");
            if(fieldName) {
                var entityEditor = document.getElementById(fieldName + "EntityEditor");
                entityEditor.querySelector("input").focus();
            }
        },

        /*
         * Handle the custom event from Awesomplete that gets called after selection is complete.
         * This replaces whatever is in the input with the selected value. 
         */
        handleSelectComplete: function(e) {
            var fieldName = e.target.getAttribute("data-fieldname");
            var entityEditor = document.getElementById(fieldName + "EntityEditor");
            var entityEditorInput = document.getElementById(fieldName + "EntityEditorInput");
            impl.appendEntity(fieldName, e.text.value, entityEditor);            
            entityEditorInput.value = "";
        },

        /*
         * Construct the outer div for the entity editor. This will just be referred to
         * as the entity editor, as it encapsulates all of that markup/functionality.
         */
        appendOuterContainer: function(current) {
            var outerDiv = document.createElement("div");

            var entityEditor = document.createElement("div");
            entityEditor.setAttribute("id", current.fieldName + 'EntityEditor');
            entityEditor.setAttribute("data-fieldname", current.fieldName);
            entityEditor.className = 'csrdemos-entityeditor';
            outerDiv.appendChild(entityEditor);

            return outerDiv;
        },

        /*
         * Add a span for each entity in ctx.CurrentFieldValue.
         */
        appendInitialEntities: function(ctx, current, entityEditor) {
            // if the field has a current value, initilize the control with it
            if (ctx.CurrentFieldValue.length > 0) {
                // parse the values into an array
                var values = ctx.CurrentFieldValue.replace(/^;#/, '').
                    replace(/;#$/, '').split(';#');

                // for each value
                values.forEach(function(v) {
                    // insert a new entity
                    impl.appendEntity(current.fieldName, v, entityEditor);
                });
            }
        },
        
        /*
         * Append a single entity as a span, just in front of the input, and remove it 
         * from the source list for the autocomplete. An entity is just a stylized span 
         * with an anchor (x) to remove it. Also, if this is after post render, attach
         * handleEntityRemove to the little x anchor. Otherwise, the post render override
         * will do that for all of the entities.
         */
        appendEntity: function(fieldName, value, entityEditor) {
            // create an anchor tag to remove this entity
            var anchor = document.createElement("a");
            anchor.setAttribute("title", "Remove Entity");
            anchor.setAttribute("href", "#");
            anchor.setAttribute("data-fieldname", fieldName);
            anchor.setAttribute("data-value", value);
                    
            // add text of an X and a class with which to style it
            anchor.className = "csrdemos-remove";
            anchor.innerHTML = "x";

            // create the span from the value and the anchor
            var span = document.createElement("span");
            span.setAttribute("title", value);
            span.setAttribute("data-fieldname", fieldName);
            span.setAttribute("data-value", value);
            span.className = "csrdemos-entity";
            span.innerHTML = value;
            span.appendChild(anchor);
            
            // push the span onto entity editor children, right before the input
            // if present, and setup event receivers, otherwise just append it as this 
            // is the initial rendering and post render will setup the event receivers
            var f = impl.fieldMap[fieldName];
            var entityEditorInput = entityEditor.querySelector("input");
            if (entityEditorInput) {
                // if the input exists, then the entity is being appended after post render
                // (like an autocomplete select), so we need to insert the entity before the
                // input and do some extra stuff
                var schema = f.schema;

                // insert the entity and clear the input
                entityEditor.insertBefore(span, entityEditorInput.parentNode);        

                // clear the input, and hide it for single select choices
                entityEditorInput.value = "";
                if (schema.FieldType !== "MultiChoice") {
                    entityEditorInput.style.display = "none";
                }

                // add a remove entitity handle to the anchor
                anchor.addEventListener("click", impl.handleEntityRemove);
            }
            else {
                entityEditor.appendChild(span);
            }
            
            // remove the value from the list of potential values, so 
            // autocomplete won't allow duplicates
            var source = f.source;
            var i = source.indexOf(value);
            if (i > -1) {
                source.splice(i, 1);
            }            

            // re-initialize autocomplete with the updated source array
            if (f.awesomplete) {
                f.awesomplete.list = source;
            }
        },

        /*
         * Add the input control to the entity editor.
         */
        appendInput: function(ctx, current, entityEditor) {
            // add an input for the user to type into, this is the autocomplete input
            var entityEditorInput = document.createElement("input");
            entityEditorInput.setAttribute("id", current.fieldName + 'EntityEditorInput');
            entityEditorInput.setAttribute("name", current.fieldName + 'EntityEditorInput');
            entityEditorInput.setAttribute("type", "text");
            entityEditorInput.setAttribute("data-fieldname", current.fieldName);
            entityEditorInput.className = 'csrdemos-entityeditorinput';

            // if single entry choice, and we already have a value, hide the input
            var f = impl.fieldMap[current.fieldName];
            var schema = f.schema;
            if (schema.FieldType !== "MultiChoice" && ctx.CurrentFieldValue.length > 0) {
                entityEditorInput.style.display = "none";
            }

            // add the input to the entity editor
            entityEditor.appendChild(entityEditorInput);

            // finally, append a span as a sibling of the outer div where we'll output 
            // any validation errors
            var span = document.createElement("span");
            span.setAttribute("id", current.fieldName + 'EntityEditorError');
            span.className = 'ms-formvalidation ms-csrformvalidation';
            entityEditor.parentNode.appendChild(span);
        },
        
        /*
         * Determine if a string is in an array ignoring case.
         */
        inArrayIgnoreCase: function(value, array) {
            value = value.toLowerCase();
            array = array.map(function(v) {
                return v.toLowerCase();
            });
            return array.indexOf(value);
        },
        
        /*
         * This code checks to see if a value is already in the entity editor. If so,
         * in addition to returning true, it clears the input, so this should only be
         * called when resolving an entity (i.e. handleEntitySelect or handleKeyDown for
         * the return key).
         */
        isDuplicateValue: function(entityEditor, value) {
            var entities = entityEditor.querySelectorAll(".csrdemos-remove");
            for (var i = 0; i < entities.length; i++) {
                if (entities[i].getAttribute("data-value").toLowerCase() === value.toLowerCase()) {
                    entityEditor.querySelector("input").value = "";
                    return true;
                }
            }
            return false;
        }
    };

    /*
     * Create an empty overrides object.
     */
    var entityEditorOverrides = {
        Templates: {
            'Fields': {}
        }
    };

    /*
     * Add an overrides object for each field we want to customize.
     */
    fields.forEach(function(v) {
        entityEditorOverrides.Templates.Fields[v] = {
            'View': impl.renderDisplay,
            'DisplayForm': impl.renderDisplay,
            'NewForm': impl.renderEdit,
            'EditForm': impl.renderEdit
        };
    });

    /*
     * Add pre and post render overrides.
     */
    entityEditorOverrides.OnPreRender = impl.preRender;
    entityEditorOverrides.OnPostRender = impl.postRender;
    
    /*
     * Register my template overrides with SPClientTemplates.
     */
    if (typeof _spPageContextInfo !== 'undefined' && _spPageContextInfo !== null) {
        // MDS is enabled
        var url = (_spPageContextInfo.siteServerRelativeUrl === '/' ? "" : _spPageContextInfo.siteServerRelativeUrl) +
            '/Style%20Library/EntityEditorCSR.js';
        // register a callback to register the templates on partial page loads
        RegisterModuleInit(url.toLowerCase(), function() {
            SPClientTemplates.TemplateManager.RegisterTemplateOverrides(entityEditorOverrides);
        });
    }

    // also just register for full page loads (F5/refresh)
    SPClientTemplates.TemplateManager.RegisterTemplateOverrides(entityEditorOverrides);

    var entityEditorCss = (function() {/*
        .csrdemos-entityeditor { border: 1px solid #ababab; width: 390px; padding: 3px; background: white; }
        .csrdemos-entityeditor:hover { border: 1px solid #92c0e0; }
        input.csrdemos-entityeditorinput { width: 140px; border: none; outline: none; }
        .csrdemos-entity { display: inline-block; padding: 2px 3px 1px 5px; margin-right: 2px; margin-bottom: 1px; position: relative; background-color: #eee; border: 1px solid #333; -moz-border-radius: 7px; -webkit-border-radius: 7px; border-radius: 7px; color: #333; font: normal 11px Verdana, Sans-serif; }
        .csrdemos-remove { margin-left: 5px; color: #0072c6; }
    */}).toString().slice(15, -3);
})();
